from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors, utility_functions
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.utility_classes.zones import Zone
from highcharts_core.options.plot_options.levels import TreemapLevelOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.breadcrumbs import BreadcrumbOptions


class TreemapOptions(GenericTypeOptions):
    """General options to apply to all Treemap series types.

    A treemap displays hierarchical data using nested rectangles. The data can be laid
    out in varying ways depending on options.

    .. figure:: ../../../_static/treemap-example.png
      :alt: Treemap Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._animation_limit = None
        self._boost_blending = None
        self._boost_threshold = None
        self._border_radius = None
        self._color_axis = None
        self._color_key = None
        self._colors = None
        self._crop_threshold = None
        self._find_nearest_point_by = None
        self._get_extremes_from_all = None
        self._ignore_hidden_point = None
        self._linecap = None
        self._line_width = None
        self._negative_color = None
        self._point_interval = None
        self._point_interval_unit = None
        self._point_start = None
        self._relative_x_value = None
        self._soft_threshold = None
        self._stacking = None
        self._step = None
        self._zone_axis = None
        self._zones = None

        self._color_by_point = None
        self._color_index = None
        self._crisp = None
        self._allow_traversing_tree = None
        self._breadcrumbs = None
        self._level_is_constant = None
        self._levels = None

        self._alternate_starting_direction = None
        self._interact_by_leaf = None
        self._layout_algorithm = None
        self._layout_starting_direction = None
        self._sort_index = None

        self.animation_limit = kwargs.get('animation_limit', None)
        self.boost_blending = kwargs.get('boost_blending', None)
        self.boost_threshold = kwargs.get('boost_threshold', None)
        self.color_axis = kwargs.get('color_axis', None)
        self.color_key = kwargs.get('color_key', None)
        self.colors = kwargs.get('colors', None)
        self.crop_threshold = kwargs.get('crop_threshold', None)
        self.find_nearest_point_by = kwargs.get('find_nearest_point_by', None)
        self.get_extremes_from_all = kwargs.get('get_extremes_from_all', None)
        self.ignore_hidden_point = kwargs.get('ignore_hidden_point', None)
        self.linecap = kwargs.get('linecap', None)
        self.line_width = kwargs.get('line_width', None)
        self.negative_color = kwargs.get('negative_color', None)
        self.point_interval = kwargs.get('point_interval', None)
        self.point_interval_unit = kwargs.get('point_interval_unit', None)
        self.point_start = kwargs.get('point_start', None)
        self.relative_x_value = kwargs.get('relative_x_value', None)
        self.soft_threshold = kwargs.get('soft_threshold', None)
        self.stacking = kwargs.get('stacking', None)
        self.step = kwargs.get('step', None)
        self.zone_axis = kwargs.get('zone_axis', None)
        self.zones = kwargs.get('zones', None)

        self.color_index = kwargs.get('color_index', None)
        self.crisp = kwargs.get('crisp', None)
        self.allow_traversing_tree = kwargs.get('allow_traversing_tree', None)
        self.breadcrumbs = kwargs.get('breadcrumbs', None)
        self.color_by_point = kwargs.get('color_by_point', None)
        self.level_is_constant = kwargs.get('level_is_constant', None)
        self.levels = kwargs.get('levels', None)

        self.alternate_starting_direction = kwargs.get('alternate_starting_direction',
                                                       None)
        self.interact_by_leaf = kwargs.get('interact_by_leaf', None)
        self.layout_algorithm = kwargs.get('layout_algorithm', None)
        self.layout_starting_direction = kwargs.get('layout_starting_direction', None)
        self.sort_index = kwargs.get('sort_index', None)

        super().__init__(**kwargs)

    @property
    def allow_traversing_tree(self) -> Optional[bool]:
        """If ``True``, the user can click on a point which is a parent and zoom in on its
        children. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_traversing_tree

    @allow_traversing_tree.setter
    def allow_traversing_tree(self, value):
        if value is None:
            self._allow_traversing_tree = None
        else:
            self._allow_traversing_tree = bool(value)

    @property
    def alternate_starting_direction(self) -> Optional[bool]:
        """If ``True``, will make the treemap alternate the drawing direction between
        vertical and horizontal for each level. The next level's starting direction will
        always be the opposite of the previous. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._alternate_starting_direction

    @alternate_starting_direction.setter
    def alternate_starting_direction(self, value):
        if value is None:
            self._alternate_starting_direction = None
        else:
            self._alternate_starting_direction = bool(value)

    @property
    def animation_limit(self) -> Optional[int | float | Decimal]:
        """For some series, there is a limit that shuts down initial animation by default
        when the total number of points in the chart is too high. Defaults to
        :obj:`None <python:None>`.

        For example, for a column chart and its derivatives, animation does not run if
        there is more than 250 points totally. To disable this cap, set
        ``animation_limit`` to ``float("inf")`` (which represents infinity).

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._animation_limit

    @animation_limit.setter
    def animation_limit(self, value):
        if value == float('inf'):
            self._animation_limit = float('inf')
        else:
            self._animation_limit = validators.numeric(value,
                                                       allow_empty = True,
                                                       minimum = 0)

    @property
    def boost_blending(self) -> Optional[str]:
        """Sets the color blending in the boost module. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boost_blending

    @boost_blending.setter
    def boost_blending(self, value):
        self._boost_blending = validators.string(value, allow_empty = True)

    @property
    def boost_threshold(self) -> Optional[int]:
        """Set the point threshold for when a series should enter boost mode. Defaults to
        ``5000``.

        Setting it to e.g. 2000 will cause the series to enter boost mode when there are
        2,000 or more points in the series.

        To disable boosting on the series, set the ``boost_threshold`` to ``0``. Setting
        it to ``1`` will force boosting.

        .. note::

          The :meth:`AreaOptions.crop_threshold` also affects this setting.

          When zooming in on a series that has fewer points than the ``crop_threshold``,
          all points are rendered although outside the visible plot area, and the
          ``boost_threshold`` won't take effect.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._boost_threshold

    @boost_threshold.setter
    def boost_threshold(self, value):
        self._boost_threshold = validators.integer(value,
                                                   allow_empty = True,
                                                   minimum = 0)

    @property
    def breadcrumbs(self) -> Optional[BreadcrumbOptions]:
        """Options for the breadcrumbs, the navigation at the top leading the way up
        through the traversed levels. Defaults to :obj:`None <python:None>`.

        """
        return self._breadcrumbs

    @breadcrumbs.setter
    @class_sensitive(BreadcrumbOptions)
    def breadcrumbs(self, value):
        self._breadcrumbs = value

    @property
    def color_axis(self) -> Optional[str | int | bool]:
        """When using dual or multiple color axes, this setting defines which
        :term:`color axis` the particular series is connected to. It refers to either the
        :meth:`ColorAxis.id` or the index of the axis in the :class:`ColorAxis` array,
        with ``0`` being the first. Set this option to ``False`` to prevent a series from
        connecting to the default color axis.

        Defaults to ``0``.

        :rtype: :obj:`None <python:None>` or :class:`str <python:str>` or
          :class:`int <python:int>` or :class:`bool <python:bool>`
        """
        return self._color_axis

    @color_axis.setter
    def color_axis(self, value):
        if value is None:
            self._color_axis = None
        elif value is False:
            self._color_axis = False
        else:
            try:
                self._color_axis = validators.string(value)
            except TypeError:
                self._color_axis = validators.integer(value,
                                                      minimum = 0)

    @property
    def color_by_point(self) -> Optional[bool]:
        """When using automatic point colors pulled from the global colors or
        series-specific collections, this option determines whether the chart should
        receive one color per series (``False``) or one color per point (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._color_by_point

    @color_by_point.setter
    def color_by_point(self, value):
        if value is None:
            self._color_by_point = None
        else:
            self._color_by_point = bool(value)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def color_key(self) -> Optional[str]:
        """Determines what data value should be used to calculate point color if
        :meth:`AreaOptions.color_axis` is used.

        .. note::

          Requires to set ``min`` and ``max`` if some custom point property is used or if
          approximation for data grouping is set to ``'sum'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color_key

    @color_key.setter
    def color_key(self, value):
        self._color_key = validators.string(value, allow_empty = True)

    @property
    def colors(self) -> Optional[List[str | Gradient | Pattern]]:
        """A series-specific or series type-specific color set to apply instead of the
        global colors when :meth:`ArcDiagramOptions.color_by_point` is ``True``.

        :rtype: :class:`list <python:list>` of :class:`str <python:str>`,
          :class:`Gradient`, or :class:`Pattern` OR :obj:`None <python:None>`
        """
        return self._colors

    @colors.setter
    def colors(self, value):
        if not value:
            self._colors = None
        else:
            value = validators.iterable(value)

            self._colors = [utility_functions.validate_color(x) for x in value]

    @property
    def crisp(self) -> Optional[bool]:
        """If ``True``, each point or column edge is rounded to its nearest pixel in order
        to render sharp on screen. Defaults to ``True``.

        .. hint::

          In some cases, when there are a lot of densely packed columns, this leads to
          visible difference in column widths or distance between columns. In these cases,
          setting ``crisp`` to ``False`` may look better, even though each column is
          rendered blurry.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crisp

    @crisp.setter
    def crisp(self, value):
        if value is None:
            self._crisp = None
        else:
            self._crisp = bool(value)

    @property
    def crop_threshold(self) -> Optional[int]:
        """When the series contains less points than the crop threshold, all points are
        drawn, even if the points fall outside the visible plot area at the current zoom.
        Defaults to ``300``.

        The advantage of drawing all points (including markers and columns), is that
        animation is performed on updates. On the other hand, when the series contains
        more points than the crop threshold, the series data is cropped to only contain
        points that fall within the plot area. The advantage of cropping away invisible
        points is to increase performance on large series.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._crop_threshold

    @crop_threshold.setter
    def crop_threshold(self, value):
        self._crop_threshold = validators.integer(value,
                                                   allow_empty = True,
                                                   minimum = 0)

    @property
    def find_nearest_point_by(self) -> Optional[str]:
        """Determines whether the series should look for the nearest point in both
        dimensions or just the x-dimension when hovering the series.

        If :obj:`None <python:None>`, defaults to ``'xy'`` for scatter series and ``'x'``
        for most other series. If the data has duplicate x-values, it is recommended to
        set this to ``'xy'`` to allow hovering over all points.

        Applies only to series types using nearest neighbor search (not direct hover) for
        tooltip.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._find_nearest_point_by

    @find_nearest_point_by.setter
    def find_nearest_point_by(self, value):
        self._find_nearest_point_by = validators.string(value, allow_empty = True)

    @property
    def get_extremes_from_all(self) -> Optional[bool]:
        """If ``True``, uses the Y extremes of the total chart width or only the zoomed
        area when zooming in on parts of the X axis. By default, the Y axis adjusts to the
        min and max of the visible data.

        .. warning::

          Applies to :term:`Cartesian series` only.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._get_extremes_from_all

    @get_extremes_from_all.setter
    def get_extremes_from_all(self, value):
        if value is None:
            self._get_extremes_from_all = None
        else:
            self._get_extremes_from_all = bool(value)

    @property
    def ignore_hidden_point(self) -> Optional[bool]:
        """If ``True``, the series shall be redrawn as if the hidden points were ``null``.
        If ``False``, hidden points will not be displayed but the slice will still be
        drawn as a gap in the pie.

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._ignore_hidden_point

    @ignore_hidden_point.setter
    def ignore_hidden_point(self, value):
        if value is None:
            self._ignore_hidden_point = None
        else:
            self._ignore_hidden_point = bool(value)

    @property
    def interact_by_leaf(self) -> Optional[bool]:
        """If ``True``, the user can only interact with leaf nodes. If ``False``, the user
        can interact with leaf and parent nodes. When :obj:`None <python:None>`, if
        :meth:`TreemapOptions.allow_traversing_tree` is ``True``, will default to
        ``False``. Otherwise will default to ``True``.

        Defaults to :obj:`None <python:None>`

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._interact_by_leaf

    @interact_by_leaf.setter
    def interact_by_leaf(self, value):
        if value is None:
            self._interact_by_leaf = None
        else:
            self._interact_by_leaf = bool(value)

    @property
    def layout_algorithm(self) -> Optional[str]:
        """This option decides which algorithm is used for setting position and dimensions
        of the points. Defaults to ``'sliceAndDice'``.

        Accepts the following (case-sensitive) values:

          * ``'sliceAndDice'``
          * ``'stripes'``
          * ``'squarified'``
          * ``'strip'``

        .. note::

          You also have the ability to extend Highcharts with your own layout algorithm.
          For more information, read the
          `original JavaScript documentation <https://www.highcharts.com/docs/chart-and-series-types/treemap#add-your-own-algorithm>`__.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._layout_algorithm

    @layout_algorithm.setter
    def layout_algorithm(self, value):
        self._layout_algorithm = validators.variable_name(value, allow_empty = True)

    @property
    def layout_starting_direction(self) -> Optional[str]:
        """Defines which direction the layout algorithm will start drawing. Defaults to
        ``'vertical'``.

        Accepts:

          * ``'vertical'``
          * ``'horizontal'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._layout_starting_direction

    @layout_starting_direction.setter
    def layout_starting_direction(self, value):
        if not value:
            self._layout_starting_direction = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['vertical', 'horizontal']:
                raise errors.HighchartsError(f'layout_starting_direction expects either '
                                             f'"vertical" or "horizontal". Received: '
                                             f'{value}')

            self._layout_starting_direction = value

    @property
    def level_is_constant(self) -> Optional[bool]:
        """If ``True``, the level will be the same as the tree structure. If ``False``,
        the first level visible when drilling is considered to be level one. Defaults to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._level_is_constant

    @level_is_constant.setter
    def level_is_constant(self, value):
        if value is None:
            self._level_is_constant = None
        else:
            self._level_is_constant = bool(value)

    @property
    def levels(self) -> Optional[List[TreemapLevelOptions]]:
        """Set options on specific levels. Takes precedence over series options, but not
        node and link options.

        :rtype: :obj:`None <python:None>`, or :class:`list <python:list>` of
          :class:`TreemapLevelOptions`
        """
        return self._levels

    @levels.setter
    @class_sensitive(TreemapLevelOptions, force_iterable = True)
    def levels(self, value):
        self._levels = value

    @property
    def linecap(self) -> Optional[str]:
        """The SVG value used for the ``stroke-linecap`` and ``stroke-linejoin`` of a line
        graph. Defaults to ``'round'``, which means that lines are rounded in the ends and
        bends.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._linecap

    @linecap.setter
    def linecap(self, value):
        self._linecap = validators.string(value, allow_empty = True)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """Pixel width of the graph line. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def negative_color(self) -> Optional[str | Gradient | Pattern]:
        """The color for the parts of the graph or points that are below the
        :meth:`AreaOptions.threshold`.

        .. note::

          :meth:`Zones <AreaOptions.zones>` take precedence over the negative color.
          Using ``negative_color`` is equivalent to applying a zone with value of 0.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._negative_color

    @negative_color.setter
    def negative_color(self, value):
        from highcharts_core import utility_functions
        self._negative_color = utility_functions.validate_color(value)

    @property
    def point_interval(self) -> Optional[int | float | Decimal]:
        """If no x values are given for the points in a series, ``point_interval`` defines
        the interval of the x values. Defaults to ``1``.

        For example, if a series contains one value every decade starting from year 0, set
        ``point_interval`` to ``10``. In true datetime axes, the ``point_interval`` is set
        in milliseconds.

        .. hint::

          ``point_interval`` can be also be combined with
          :meth:`point_interval_unit <AreaOptions.point_interval_unit>` to draw irregular
          time intervals.

        .. note::

          If combined with :meth:`relative_x_value <AreaOptions.relative_x_value>`, an x
          value can be set on each point, and the ``point_interval`` is added x times to
          the :meth:`point_start <AreaOptions.point_start>` setting.

        .. warning::

          This options applies to the series data, not the interval of the axis ticks,
          which is independent.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_interval

    @point_interval.setter
    def point_interval(self, value):
        self._point_interval = validators.numeric(value,
                                                  allow_empty = True,
                                                  minimum = 0)

    @property
    def point_interval_unit(self) -> Optional[str]:
        """On datetime series, this allows for setting the
        :meth:`point_interval <AreaOptions.point_interval>` to irregular time units, day,
        month, and year.

        A day is usually the same as 24 hours, but ``point_interval_unit`` also takes the
        DST crossover into consideration when dealing with local time.

        Combine this option with :meth:`point_interval <AreaOptions.point_interval>` to
        draw weeks, quarters, 6 month periods, 10 year periods, etc.

        .. warning::

          This options applies to the series data, not the interval of the axis ticks,
          which is independent.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._point_interval_unit

    @point_interval_unit.setter
    def point_interval_unit(self, value):
        self._point_interval_unit = validators.string(value, allow_empty = True)

    @property
    def point_start(self) -> Optional[int | float | Decimal]:
        """If no x values are given for the points in a series, ``point_start`` defines
        on what value to start. For example, if a series contains one yearly value
        starting from 1945, set ``point_start`` to ``1945``. Defaults to ``0``.

        .. note::

          If combined with :meth:`relative_x_value <AreaOptions.relative_x_value>`, an x
          value can be set on each point. The x value from the point options is multiplied
          by :meth:`point_interval <AreaOptions.point_interval>` and added to
          ``point_start`` to produce a modified x value.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_start

    @point_start.setter
    def point_start(self, value):
        self._point_start = validators.numeric(value, allow_empty = True)

    @property
    def relative_x_value(self) -> Optional[bool]:
        """When ``True``, X values in the data set are relative to the current
        :meth:`point_start <AreaOptions.point_start>`,
        :meth:`point_interval <AreaOptions.point_interval>`, and
        :meth:`point_interval_unit <AreaOptions.point_interval_unit>` settings. This
        allows compression of the data for datasets with irregular X values. Defaults to
        ``False``.

        The real X values are computed on the formula ``f(x) = ax + b``, where ``a`` is
        the :meth:`point_interval <AreaOptions.point_interval>` (optionally with a time
        unit given by :meth:`point_interval_unit <AreaOptions.point_interval_unit>`), and
        ``b`` is the :meth:`point_start <AreaOptions.point_start>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._relative_x_value

    @relative_x_value.setter
    def relative_x_value(self, value):
        if value is None:
            self._relative_x_value = None
        else:
            self._relative_x_value = bool(value)

    @property
    def soft_threshold(self) -> Optional[bool]:
        """When ``True``, the series will not cause the Y axis to cross the zero plane (or
        threshold option) unless the data actually crosses the plane. Defaults to
        ``True``.

        For example, if ``False``, a series of ``0, 1, 2, 3`` will make the Y axis show
        negative values according to the ``min_padidng`` option. If ``True``, the Y axis
        starts at 0.

        :rtype: :class:`bool <python:bool>`
        """
        return self._soft_threshold

    @soft_threshold.setter
    def soft_threshold(self, value):
        if value is None:
            self._soft_threshold = None
        else:
            self._soft_threshold = bool(value)

    @property
    def sort_index(self) -> Optional[int]:
        """The sort index of the point inside the treemap level. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._sort_index

    @sort_index.setter
    def sort_index(self, value):
        self._sort_index = validators.integer(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def stacking(self) -> Optional[str]:
        """Whether to stack the values of each series on top of each other. Defaults to
        :obj:`None <python:None>`.

        Acceptable values are:

          * :obj:`None <python:None>` to disable stacking,
          * ``"normal"`` to stack by value or
          * ``"percent"``
          * ``'stream'`` (for streamgraph series type only)
          * ``'overlap'`` (for waterfall series type only)

        .. note::

          When stacking is enabled, data must be sorted in ascending X order.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stacking

    @stacking.setter
    def stacking(self, value):
        if not value:
            self._stacking = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['normal', 'percent', 'stream', 'overlap']:
                raise errors.HighchartsValueError(f'stacking expects a valid stacking '
                                                  f'value. However, received: {value}')
            self._stacking = value

    @property
    def step(self) -> Optional[str]:
        """Whether to apply steps to the line. Defaults to :obj:`None <python:None>`.

        Possible values are:

          * :obj:`None <python:None>`
          * ``'left'``
          * ``'center'``
          * ``'right'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._step

    @step.setter
    def step(self, value):
        self._step = validators.string(value, allow_empty = True)

    @property
    def zone_axis(self) -> Optional[str]:
        """Defines the Axis on which the zones are applied. Defaults to ``'y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._zone_axis

    @zone_axis.setter
    def zone_axis(self, value):
        self._zone_axis = validators.string(value, allow_empty = True)

    @property
    def zones(self) -> Optional[List[Zone]]:
        """An array defining zones within a series. Defaults to :obj:`None <python:None>`.

        Zones can be applied to the X axis, Y axis or Z axis for bubbles, according to the
        :meth:`zone_axis <AreaOptions.zone_axis>` setting.

        .. warning::

          The zone definitions have to be in ascending order regarding to the value.

        :rtype: :obj:`None <python:None>` or :class:`list <python:list>` of
          :class:`Zone` instances
        """
        return self._zones

    @zones.setter
    @class_sensitive(Zone, force_iterable = True)
    def zones(self, value):
        self._zones = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_key': as_dict.get('colorKey', None),
            'colors': as_dict.get('colors', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'ignore_hidden_point': as_dict.get('ignoreHiddenPoint', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'color_index': as_dict.get('colorIndex', None),
            'crisp': as_dict.get('crisp', None),
            'allow_traversing_tree': as_dict.get('allowTraversingTree', None),
            'breadcrumbs': as_dict.get('breadcrumbs', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'level_is_constant': as_dict.get('levelIsConstant', None),
            'levels': as_dict.get('levels', None),

            'alternate_starting_direction': as_dict.get('alternateStartingDirection',
                                                        None),
            'interact_by_leaf': as_dict.get('interactByLeaf', None),
            'layout_algorithm': as_dict.get('layoutAlgorithm', None),
            'layout_starting_direction': as_dict.get('layoutStartingDirection', None),
            'sort_index': as_dict.get('sortIndex', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'allowTraversingTree': self.allow_traversing_tree,
            'alternateStartingDirection': self.alternate_starting_direction,
            'animationLimit': self.animation_limit,
            'boostBlending': self.boost_blending,
            'boostThreshold': self.boost_threshold,
            'breadcrumbs': self.breadcrumbs,
            'colorAxis': self.color_axis,
            'colorByPoint': self.color_by_point,
            'colorIndex': self.color_index,
            'colorKey': self.color_key,
            'colors': self.colors,
            'crisp': self.crisp,
            'cropThreshold': self.crop_threshold,
            'findNearestPointBy': self.find_nearest_point_by,
            'getExtremesFromAll': self.get_extremes_from_all,
            'ignoreHiddenPoint': self.ignore_hidden_point,
            'interactByLeaf': self.interact_by_leaf,
            'layoutAlgorithm': self.layout_algorithm,
            'layoutStartingDirection': self.layout_starting_direction,
            'levelIsConstant': self.level_is_constant,
            'levels': self.levels,
            'linecap': self.linecap,
            'lineWidth': self.line_width,
            'negativeColor': self.negative_color,
            'pointInterval': self.point_interval,
            'pointIntervalUnit': self.point_interval_unit,
            'pointStart': self.point_start,
            'relativeXValue': self.relative_x_value,
            'softThreshold': self.soft_threshold,
            'sortIndex': self.sort_index,
            'stacking': self.stacking,
            'step': self.step,
            'zoneAxis': self.zone_axis,
            'zones': self.zones
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class ScrollablePlotArea(HighchartsMeta):
    """Configuration settings to make the plot area scrollable.

    This feature provides a minimum size for the plot area of the chart. If the size
    gets smaller than this, typically on mobile devices, a native browser scrollbar is
    presented. This scrollbar provides smooth scrolling for the contents of the plot
    area, whereas the title, legend and unaffected axes are fixed.

    .. hint::

      Since v7.1.2, a scrollable plot area can be defined for either horizontal or
      vertical scrolling, depending on whether the `minimum_width` or `minimum_height`
      options are set.

    """

    def __init__(self, **kwargs):
        self._minimum_height = None
        self._minimum_width = None
        self._opacity = None
        self._scroll_position_x = None
        self._scroll_position_y = None

        self.minimum_height = kwargs.get('minimum_height', None)
        self.minimum_width = kwargs.get('minimum_width', None)
        self.opacity = kwargs.get('opacity', None)
        self.scroll_position_x = kwargs.get('scroll_position_x', None)
        self.scroll_position_y = kwargs.get('scroll_position_y', None)

    @property
    def minimum_height(self) -> Optional[int | float | Decimal]:
        """The minimum height for the plot area expressed in pixels. If it gets smaller
        than this, the plot area will become scrollable.

        :rtype: numeric
        """
        return self._minimum_height

    @minimum_height.setter
    def minimum_height(self, value):
        self._minimum_height = validators.numeric(value, allow_empty = True)

    @property
    def minimum_width(self) -> Optional[int | float | Decimal]:
        """The minimum width for the plot area expressed in pixels. If it gets smaller
        than this, the plot area will become scrollable.

        :rtype: numeric
        """
        return self._minimum_width

    @minimum_width.setter
    def minimum_width(self, value):
        self._minimum_width = validators.numeric(value, allow_empty = True)

    @property
    def opacity(self) -> Optional[float]:
        """The opacity of the mask applied on one of the sides of the plot area, expressed
        as a value between ``0`` and ``1``. Defaults to ``0.85``.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        self._opacity = validators.float(value, allow_empty = True)

    @property
    def scroll_position_x(self) -> Optional[float]:
        """The initial scrolling position of the scrollable plot area along the
        horizontal axis. Ranges from 0 to 1, where 0 aligns the plot area to the left and
        1 aligns it to the right.

        .. hint::

          Typically, we would use a value of ``1`` if the chart has right aligned Y axes.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._scroll_position_x

    @scroll_position_x.setter
    def scroll_position_x(self, value):
        self._scroll_position_x = validators.numeric(value,
                                                     allow_empty = True,
                                                     minimum = 0,
                                                     maximum = 1)

    @property
    def scroll_position_y(self) -> Optional[float]:
        """The initial scrolling position of the scrollable plot area along the
        vertical axis. Ranges from 0 to 1, where 0 aligns the plot area to the top and
        1 aligns it to the bottom.

        .. hint::

          Typically, we would use a value of ``1`` if the chart has right aligned Y axes.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._scroll_position_y

    @scroll_position_y.setter
    def scroll_position_y(self, value):
        self._scroll_position_y = validators.numeric(value,
                                                     allow_empty = True,
                                                     minimum = 0,
                                                     maximum = 1)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'minimum_height': as_dict.get('minHeight', None),
            'minimum_width': as_dict.get('minWidth', None),
            'opacity': as_dict.get('opacity', None),
            'scroll_position_x': as_dict.get('scrollPositionX', None),
            'scroll_position_y': as_dict.get('scrollPositionY', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'minHeight': self.minimum_height,
            'minWidth': self.minimum_width,
            'opacity': self.opacity,
            'scrollPositionX': self.scroll_position_x,
            'scrollPositionY': self.scroll_position_y
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.plot_options.pie import PieOptions as cls
from highcharts_core.options.plot_options.pie import VariablePieOptions as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    # Pie Options only
    ({
      'border_color': '#ccc',
      'border_width': 1,
      'center': ['50%', '50%'],
      'color_axis': 1,
      'color_index': 3,
      'color_key': 'some-key-goes-here',
      'colors': [
          '#fff',
          'ccc',
          {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.567,
                'y1': 0.891,
                'y2': 0.987
            },
            'stops': [
                [0.123, '#cccccc'],
                [0.456, '#ff0000'],
                [1, '#00ff00']
            ]
          },
          {
            'animation': {
                'defer': 5
            },
            'patternOptions': {
                'aspectRatio': 0.5,
                'backgroundColor': '#999999',
                'id': 'some_id_goes_here',
                'opacity': 0.5,
                'width': 120,
                'x': 5,
                'y': 10
            },
            'patternIndex': 2
          }
      ],
      'depth': 10,
      'end_angle': 90,
      'fill_color': '#fff',
      'ignore_hidden_point': True,
      'inner_size': '30%',
      'linecap': 'round',
      'min_size': '20%',
      'size': 80,
      'sliced_offset': 24,
      'start_angle': 45,
      'thickness': 2
    }, None),
    # + Generic Options
    ({
      'border_color': '#ccc',
      'border_width': 1,
      'center': ['50%', '50%'],
      'color_axis': 1,
      'color_index': 3,
      'color_key': 'some-key-goes-here',
      'colors': [
          '#fff',
          'ccc',
          {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.567,
                'y1': 0.891,
                'y2': 0.987
            },
            'stops': [
                [0.123, '#cccccc'],
                [0.456, '#ff0000'],
                [1, '#00ff00']
            ]
          },
          {
            'animation': {
                'defer': 5
            },
            'patternOptions': {
                'aspectRatio': 0.5,
                'backgroundColor': '#999999',
                'id': 'some_id_goes_here',
                'opacity': 0.5,
                'width': 120,
                'x': 5,
                'y': 10
            },
            'patternIndex': 2
          }
      ],
      'depth': 10,
      'end_angle': 90,
      'fill_color': '#fff',
      'ignore_hidden_point': True,
      'inner_size': '30%',
      'linecap': 'round',
      'min_size': '20%',
      'size': 80,
      'sliced_offset': 24,
      'start_angle': 45,
      'thickness': 2,

      'accessibility': {
          'description': 'Description goes here',
          'enabled': True,
          'exposeAsGroupOnly': True,
          'keyboardNavigation': {
              'enabled': True
          },
          'point': {
              'dateFormat': 'format string',
              'dateFormatter': """function() { return true; }""",
              'describeNull': False,
              'descriptionFormatter': """function() { return true; }""",
              'valueDecimals': 2,
              'valueDescriptionFormat': 'format string',
              'valuePrefix': '$',
              'valueSuffix': 'USD'
          },
      },
      'allow_point_select': True,
      'animation': {
          'defer': 5
      },
      'class_name': 'some-class-name',
      'clip': False,
      'color': '#fff',
      'cursor': 'alias',
      'custom': {
          'item1': 'some value',
          'item2': 'some value'
      },
      'dash_style': 'Dash',
      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'description': 'Description goes here',
      'enable_mouse_tracking': True,
      'events': {
        'afterAnimate': """function(event) { return true; }""",
        'click': """function(event) { return true; }""",
        'hide': """function(event) { return true; }""",
        'mouseOut': """function(event) { return true; }""",
        'show': """function(event) { return true; }"""
      },
      'include_in_data_export': True,
      'keys': [
          'somevalue',
          'somevalue',
          'somevalue'
      ],
      'label': {
          'boxesToAvoid': [
              {
               'bottom': 12,
               'left': -46,
               'right': 84,
               'top': 24
              },
              {
               'bottom': 48,
               'left': -46,
               'right': 84,
               'top': 86
              }
          ],
          'connectorAllowed': True,
          'connectorNeighbourDistance': 12,
          'enabled': True,
          'format': 'format string',
          'formatter': """function() { return true; }""",
          'maxFontSize': 18,
          'minFontSize': 6,
          'onArea': False,
          'style': 'some style string'
      },
      'linked_to': 'some_id',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'on_point': {
          'connectorOptions': {
              'dashstyle': 'Dash',
              'stroke': '#ccc',
              'width': 2
          },
          'id': 'some-id',
          'position': {
              'align': 'left',
              'verticalAlign': 'top',
              'x': 15,
              'y': -46
          }
      },
      'opacity': 0.2,
      'point': {
          'events': {
            'click': """function(event) { return true; }""",
            'drag': """function(event) { return true; }""",
            'drop': """function(event) { return true; }""",
            'mouseOut': """function(event) { return true; }"""
          }
      },
      'point_description_formatter': """function (point) { return true; }""",
      'selected': False,
      'show_checkbox': True,
      'show_in_legend': True,
      'skip_keyboard_navigation': False,
      'states': {
        'hover': {
            'animation': {
                'duration': 123
            },
            'borderColor': '#cccccc',
            'brightness': 0.3,
            'enabled': True
        },
        'inactive': {
            'enabled': True,
            'opacity': 0.5
        },
        'normal': {
            'animation': {
                'defer': 24
            }
        },
        'select': {
            'color': '#ff0000',
            'enabled': True,
        }
      },
      'sticky_tracking': True,
      'threshold': 123,
      'tooltip': {
        'animation': True,
        'backgroundColor': '#ccc',
        'borderColor': '#999',
        'borderRadius': 4,
        'borderWidth': 1,
        'className': 'some-class-name',
        'clusterFormat': 'format string',
        'dateTimeLabelFormats': {
          'day': 'test',
          'hour': 'test',
          'millisecond': 'test',
          'minute': 'test',
          'month': 'test',
          'second': 'test',
          'week': 'test',
          'year': 'test'
        },
        'distance': 12,
        'enabled': True,
        'followPointer': True,
        'followTouch_move': True,
        'footerFormat': 'format string',
        'formatter': """function() { return true; }""",
        'headerFormat': 'format string',
        'headerShape': 'circle',
        'hideDelay': 3,
        'nullFormat': 'format string',
        'nullFormatter': """function() { return true; }""",
        'outside': False,
        'padding': 6,
        'pointFormat': 'format string',
        'pointFormatter': """function() { return true; }""",
        'positioner': """function() { return true; }""",
        'shadow': False,
        'shape': 'rect',
        'shared': False,
        'snap': 4,
        'split': False,
        'stickOnContact': True,
        'style': 'style string goes here',
        'useHTML': False,
        'valueDecimals': 2,
        'valuePrefix': '$',
        'valueSuffix': ' USD',
        'xDateFormat': 'format string'
       },
      'turbo_threshold': 456,
      'visible': True
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_PieOptions__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_PieOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_PieOptions_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_PieOptions_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('plot_options/pie/01.js', False, None),
    ('plot_options/pie/02.js', False, None),

    ('plot_options/pie/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('plot_options/pie/error-02.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('plot_options/pie/01.js', True, None),
    ('plot_options/pie/02.js', True, None),

    ('plot_options/pie/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('plot_options/pie/error-02.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_PieOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


## NEXT CLASS

STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'max_point_size': '15%',
      'min_point_size': '5%',
      'size_by': 'area',
      'z_max': 30,
      'z_min': 2
    }, None),
    # + Pie Options
    ({
      'max_point_size': '15%',
      'min_point_size': '5%',
      'size_by': 'area',
      'z_max': 30,
      'z_min': 2,

      'border_color': '#ccc',
      'border_width': 1,
      'center': ['50%', '50%'],
      'color_axis': 1,
      'color_index': 3,
      'color_key': 'some-key-goes-here',
      'colors': [
          '#fff',
          'ccc',
          {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.567,
                'y1': 0.891,
                'y2': 0.987
            },
            'stops': [
                [0.123, '#cccccc'],
                [0.456, '#ff0000'],
                [1, '#00ff00']
            ]
          },
          {
            'animation': {
                'defer': 5
            },
            'patternOptions': {
                'aspectRatio': 0.5,
                'backgroundColor': '#999999',
                'id': 'some_id_goes_here',
                'opacity': 0.5,
                'width': 120,
                'x': 5,
                'y': 10
            },
            'patternIndex': 2
          }
      ],
      'depth': 10,
      'end_angle': 90,
      'fill_color': '#fff',
      'ignore_hidden_point': True,
      'inner_size': '30%',
      'linecap': 'round',
      'min_size': '20%',
      'size': 80,
      'sliced_offset': 24,
      'start_angle': 45,
      'thickness': 2
    }, None),
    # + Generic Options
    ({
      'max_point_size': '15%',
      'min_point_size': '5%',
      'size_by': 'area',
      'z_max': 30,
      'z_min': 2,

      'border_color': '#ccc',
      'border_width': 1,
      'center': ['50%', '50%'],
      'color_axis': 1,
      'color_index': 3,
      'color_key': 'some-key-goes-here',
      'colors': [
          '#fff',
          'ccc',
          {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.567,
                'y1': 0.891,
                'y2': 0.987
            },
            'stops': [
                [0.123, '#cccccc'],
                [0.456, '#ff0000'],
                [1, '#00ff00']
            ]
          },
          {
            'animation': {
                'defer': 5
            },
            'patternOptions': {
                'aspectRatio': 0.5,
                'backgroundColor': '#999999',
                'id': 'some_id_goes_here',
                'opacity': 0.5,
                'width': 120,
                'x': 5,
                'y': 10
            },
            'patternIndex': 2
          }
      ],
      'depth': 10,
      'end_angle': 90,
      'fill_color': '#fff',
      'ignore_hidden_point': True,
      'inner_size': '30%',
      'linecap': 'round',
      'min_size': '20%',
      'size': 80,
      'sliced_offset': 24,
      'start_angle': 45,
      'thickness': 2,

      'accessibility': {
          'description': 'Description goes here',
          'enabled': True,
          'exposeAsGroupOnly': True,
          'keyboardNavigation': {
              'enabled': True
          },
          'point': {
              'dateFormat': 'format string',
              'dateFormatter': """function() { return true; }""",
              'describeNull': False,
              'descriptionFormatter': """function() { return true; }""",
              'valueDecimals': 2,
              'valueDescriptionFormat': 'format string',
              'valuePrefix': '$',
              'valueSuffix': 'USD'
          },
      },
      'allow_point_select': True,
      'animation': {
          'defer': 5
      },
      'class_name': 'some-class-name',
      'clip': False,
      'color': '#fff',
      'cursor': 'alias',
      'custom': {
          'item1': 'some value',
          'item2': 'some value'
      },
      'dash_style': 'Dash',
      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'description': 'Description goes here',
      'enable_mouse_tracking': True,
      'events': {
        'afterAnimate': """function(event) { return true; }""",
        'click': """function(event) { return true; }""",
        'hide': """function(event) { return true; }""",
        'mouseOut': """function(event) { return true; }""",
        'show': """function(event) { return true; }"""
      },
      'include_in_data_export': True,
      'keys': [
          'somevalue',
          'somevalue',
          'somevalue'
      ],
      'label': {
          'boxesToAvoid': [
              {
               'bottom': 12,
               'left': -46,
               'right': 84,
               'top': 24
              },
              {
               'bottom': 48,
               'left': -46,
               'right': 84,
               'top': 86
              }
          ],
          'connectorAllowed': True,
          'connectorNeighbourDistance': 12,
          'enabled': True,
          'format': 'format string',
          'formatter': """function() { return true; }""",
          'maxFontSize': 18,
          'minFontSize': 6,
          'onArea': False,
          'style': 'some style string'
      },
      'linked_to': 'some_id',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'on_point': {
          'connectorOptions': {
              'dashstyle': 'Dash',
              'stroke': '#ccc',
              'width': 2
          },
          'id': 'some-id',
          'position': {
              'align': 'left',
              'verticalAlign': 'top',
              'x': 15,
              'y': -46
          }
      },
      'opacity': 0.2,
      'point': {
          'events': {
            'click': """function(event) { return true; }""",
            'drag': """function(event) { return true; }""",
            'drop': """function(event) { return true; }""",
            'mouseOut': """function(event) { return true; }"""
          }
      },
      'point_description_formatter': """function (point) { return true; }""",
      'selected': False,
      'show_checkbox': True,
      'show_in_legend': True,
      'skip_keyboard_navigation': False,
      'states': {
        'hover': {
            'animation': {
                'duration': 123
            },
            'borderColor': '#cccccc',
            'brightness': 0.3,
            'enabled': True
        },
        'inactive': {
            'enabled': True,
            'opacity': 0.5
        },
        'normal': {
            'animation': {
                'defer': 24
            }
        },
        'select': {
            'color': '#ff0000',
            'enabled': True,
        }
      },
      'sticky_tracking': True,
      'threshold': 123,
      'tooltip': {
        'animation': True,
        'backgroundColor': '#ccc',
        'borderColor': '#999',
        'borderRadius': 4,
        'borderWidth': 1,
        'className': 'some-class-name',
        'clusterFormat': 'format string',
        'dateTimeLabelFormats': {
          'day': 'test',
          'hour': 'test',
          'millisecond': 'test',
          'minute': 'test',
          'month': 'test',
          'second': 'test',
          'week': 'test',
          'year': 'test'
        },
        'distance': 12,
        'enabled': True,
        'followPointer': True,
        'followTouch_move': True,
        'footerFormat': 'format string',
        'formatter': """function() { return true; }""",
        'headerFormat': 'format string',
        'headerShape': 'circle',
        'hideDelay': 3,
        'nullFormat': 'format string',
        'nullFormatter': """function() { return true; }""",
        'outside': False,
        'padding': 6,
        'pointFormat': 'format string',
        'pointFormatter': """function() { return true; }""",
        'positioner': """function() { return true; }""",
        'shadow': False,
        'shape': 'rect',
        'shared': False,
        'snap': 4,
        'split': False,
        'stickOnContact': True,
        'style': 'style string goes here',
        'useHTML': False,
        'valueDecimals': 2,
        'valuePrefix': '$',
        'valueSuffix': ' USD',
        'xDateFormat': 'format string'
       },
      'turbo_threshold': 456,
      'visible': True
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_VariablePieOptions__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_VariablePieOptions__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_VariablePieOptions_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_VariablePieOptions_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('plot_options/pie/03.js', False, None),
    ('plot_options/pie/04.js', False, None),

    ('plot_options/pie/error-03.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('plot_options/pie/error-04.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('plot_options/pie/03.js', True, None),
    ('plot_options/pie/04.js', True, None),

    ('plot_options/pie/error-03.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
    ('plot_options/pie/error-04.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_VariablePieOptions_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal
from collections import UserDict

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.series.data.base import DataBase
from highcharts_core.options.plot_options.drag_drop import DragDropOptions
from highcharts_core.utility_classes.data_labels import DataLabel


class SinglePointBase(DataBase):
    """Data point that features a single and ``y`` value."""

    def __init__(self, **kwargs):
        self._data_labels = None
        self._drag_drop = None
        self._drilldown = None

        self.data_labels = kwargs.get('data_labels', None)
        self.drag_drop = kwargs.get('drag_drop', None)
        self.drilldown = kwargs.get('drilldown', None)

        super().__init__(**kwargs)

    @property
    def data_labels(self) -> Optional[DataLabel]:
        """Individual data label for the data point.

        :rtype: :class:`DataLabel` or :obj:`None <python:None>`
        """
        return self._data_labels

    @data_labels.setter
    @class_sensitive(DataLabel)
    def data_labels(self, value):
        self._data_labels = value

    @property
    def drag_drop(self) -> Optional[DragDropOptions]:
        """The draggable-points module allows points to be moved around or modified in the
        chart.

        In addition to the options mentioned under the dragDrop API structure, the module
        fires three (JavaScript) events:

          * ``point.dragStart``
          * ``point.drag``
          * ``point.drop``

        :rtype: :class:`DragDropOptions` or :obj:`None <python:None>`
        """
        return self._drag_drop

    @drag_drop.setter
    @class_sensitive(DragDropOptions)
    def drag_drop(self, value):
        self._drag_drop = value

    @property
    def drilldown(self) -> Optional[str]:
        """The :meth:`id <SeriesBase.id>` of a series in the ``drilldown.series`` array
        to use as a drilldown destination for this point. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._drilldown

    @drilldown.setter
    def drilldown(self, value):
        if isinstance(value, constants.EnforcedNullType):
            self._drilldown = value
        else:
            self._drilldown = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,

        }

        return untrimmed


class SinglePointData(SinglePointBase):
    """Data point that features a single and ``y`` value."""

    def __init__(self, **kwargs):
        self._y = None

        self.y = kwargs.get('y', None)

        super().__init__(**kwargs)

    @property
    def y(self) -> Optional[int | float | Decimal | type(None) | constants.EnforcedNullType]:
        """The position of the data point on the Y-axis. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        if value is None or isinstance(value, constants.EnforcedNullType):
            self._y = None
        else:
            self._y = validators.numeric(value)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'SinglePointData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(y = None)
            elif checkers.is_numeric(item):
                as_obj = cls(y = item)
            elif checkers.is_iterable(item, forbid_literals = (str, bytes, dict, UserDict)):
                if len(item) == 2:
                    as_obj = cls(name = item[0],
                                 y = item[1])
                elif len(item) == 1:
                    as_obj = cls(y = item[0])
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Single Point Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),

            'y': as_dict.get('y', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'y': self.y,

            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,

        }

        return untrimmed


class SingleValueData(SinglePointBase):
    """Data point that features a single and ``value`` value."""

    def __init__(self, **kwargs):
        self._value = None

        self.value = kwargs.get('value', None)

        super().__init__(**kwargs)

    @property
    def value(self) -> Optional[int | float | Decimal | type(None) | constants.EnforcedNullType]:
        """The value of the data point. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value_):
        if value_ is None or isinstance(value_, constants.EnforcedNullType):
            self._value = None
        else:
            self._value = validators.numeric(value_)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'SingleValueData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(value = None)
            elif checkers.is_numeric(item):
                as_obj = cls(value = item)
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Single Value Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'value': as_dict.get('value', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'value': self.value,

            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,

        }

        return untrimmed


class SingleXData(SinglePointBase):
    """Data point that features a single labeled ``x`` value."""

    def __init__(self, **kwargs):
        self._x = None

        self.x = kwargs.get('x', None)

        super().__init__(**kwargs)

    @property
    def x(self) -> Optional[int | float | Decimal | type(None) | constants.EnforcedNullType]:
        """The position of the data point on the X-axis. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        if value is None or isinstance(value, constants.EnforcedNullType):
            self._x = None
        else:
            self._x = validators.numeric(value)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'SinglePointData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(x = None)
            elif checkers.is_numeric(item):
                as_obj = cls(x = item)
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Single X-value Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),

            'x': as_dict.get('x', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'x': self.x,

            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,

        }

        return untrimmed


class LabeledSingleXData(SingleXData):
    """Data point that features a single labeled ``x`` value."""

    def __init__(self, **kwargs):
        self._label = None

        self.label = kwargs.get('label', None)

        super().__init__(**kwargs)

    @property
    def label(self) -> Optional[str]:
        """The label for the event. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._label

    @label.setter
    def label(self, value):
        self._label = validators.string(value, allow_empty = True)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'LabeledSingleXData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(x = None)
            elif checkers.is_numeric(item):
                as_obj = cls(x = item)
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Single X-value Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),

            'label': as_dict.get('label', None),
            'x': as_dict.get('x', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'label': self.label,
            'x': self.x,

            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,


        }

        return untrimmed


class ConnectedSingleXData(SingleXData):
    """Variant of :class:`SingleXData` which supports a connector."""

    def __init__(self, **kwargs):
        self._connector_color = None
        self._connector_width = None

        self.connector_color = kwargs.get('connector_color', None)
        self.connector_width = kwargs.get('connector_width', None)

        super().__init__(**kwargs)

    @property
    def connector_color(self) -> Optional[str]:
        """Color of the line that connects the dumbbell point's values. If
        :obj:`None <python:None>`, applies the series' color. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._connector_color

    @connector_color.setter
    def connector_color(self, value):
        self._connector_color = validators.string(value, allow_empty = True)

    @property
    def connector_width(self) -> Optional[int | float | Decimal]:
        """Pixel width of the line that connects the dumbbell point's values. Defaults to
        ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._connector_width

    @connector_width.setter
    def connector_width(self, value):
        self._connector_width = validators.numeric(value, allow_empty = True)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'ConnectedSingleXData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(x = None)
            elif checkers.is_numeric(item):
                as_obj = cls(x = item)
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Single X-value Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank',
                                      None) or as_dict.get('labelRank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),

            'x': as_dict.get('x', None),

            'connector_color': as_dict.get('connectorColor', None),
            'connector_width': as_dict.get('connectorWidth', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'connectorColor': self.connector_color,
            'connectorWidth': self.connector_width,

            'x': self.x,

            'dataLabels': self.data_labels,
            'dragDrop': self.drag_drop,
            'drilldown': self.drilldown,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.chart import ChartOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'align_thresholds': True,
      'align_ticks': True,
      'allow_mutating_data': True,
      'animation': False,
      'background_color': '#fff',
      'border_color': '#ccc',
      'border_radius': 3,
      'border_width': 1,
      'class_name': 'some-class-name',
      'color_count': 10,
      'display_errors': True,
      'events': {
        'addSeries': """function(event) { return true;}""",
        'afterPrint': """function(event) {return true;}""",
        'click': """function(event) { return true; }""",
        'selection': """function(event) { return true; }"""
      },
      'height': 120,
      'ignore_hidden_series': False,
      'inverted': False,
      'margin': 20,
      'number_formatter': """function(value) { return true; }""",
      'pan_key': 'ctrl',
      'panning': {
          'enabled': True,
          'type': 'x'
      },
      'parallel_coordinates': False,
      'plot_background_color': '#ccc',
      'plot_background_image': 'http://www.somewhere.com',
      'plot_border_color': '#999',
      'plot_border_width': 1,
      'plot_shadow': False,
      'polar': False,
      'reflow': False,
      'render_to': 'some-id',
      'scrollable_plot_area': {
          'minHeight': 120,
          'minWidth': 300,
          'opacity': 0.6,
          'scrollPositionX': 0,
          'scrollPositionY': 0
      },
      'selection_marker_fill': '#ccc',
      'shadow': False,
      'show_axes': True,
      'spacing': [5, 5, 5, 5],
      'style': 'style-string-goes-here',
      'styled_mode': False,
      'type': 'line',
      'width': 50
    }, None),
    ({
      'align_thresholds': True,
      'align_ticks': True,
      'allow_mutating_data': True,
      'animation': False,
      'background_color': '#fff',
      'border_color': '#ccc',
      'border_radius': 3,
      'border_width': 1,
      'class_name': 'some-class-name',
      'color_count': 10,
      'display_errors': True,
      'events': {
        'addSeries': """function(event) { return true;}""",
        'afterPrint': """function(event) {return true;}""",
        'click': """function(event) { return true; }""",
        'selection': """function(event) { return true; }"""
      },
      'height': 120,
      'ignore_hidden_series': False,
      'inverted': False,
      'margin': 20,
      'number_formatter': """function(value) { return true; }""",
      'pan_key': 'ctrl',
      'panning': {
          'enabled': True,
          'type': 'x'
      },
      'parallel_coordinates': False,
      'plot_background_color': '#ccc',
      'plot_background_image': 'http://www.somewhere.com',
      'plot_border_color': '#999',
      'plot_border_width': 1,
      'plot_shadow': False,
      'polar': False,
      'reflow': False,
      'render_to': 'some-id',
      'scrollable_plot_area': {
          'minHeight': 120,
          'minWidth': 300,
          'opacity': 0.6,
          'scrollPositionX': 0,
          'scrollPositionY': 0
      },
      'selection_marker_fill': '#ccc',
      'shadow': False,
      'show_axes': True,
      'spacing': [5, 5, 5, 5],
      'style': 'style-string-goes-here',
      'styled_mode': False,
      'type': 'line',
      'width': 50,
      'zooming': {
          'key': 'alt',
          'mouse_wheel': {
              'enabled': True,
              'sensitivity': 1.5,
              'type': 'xy'
          },
          'pinch_type': 'xy',
          'single_touch': False
      }
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('chart/chart/01.js', False, None),
    ('chart/chart/02.js', False, None),

    ('chart/chart/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('chart/chart/01.js', True, None),
    ('chart/chart/02.js', False, None),

    ('chart/chart/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.chart.options_3d import Options3D as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'alpha': 0.5,
      'axis_label_position': 'auto',
      'beta': 15,
      'depth': 20,
      'enabled': True,
      'fit_to_plot': True,
      'frame': {
          'back': {
              'color': '#ccc',
              'size': 2,
              'visible': 'auto'
          },
          'bottom': {
              'color': '#ccc',
              'size': 2,
              'visible': 'auto'
          },
          'front': {
              'color': '#ccc',
              'size': 2,
              'visible': 'auto'
          },
          'left': {
              'color': '#ccc',
              'size': 2,
              'visible': 'auto'
          },
          'right': {
              'color': '#ccc',
              'size': 2,
              'visible': 'auto'
          },
          'top': {
              'color': '#ccc',
              'size': 2,
              'visible': 'auto'
          },
          'size': 2,
          'visible': True
      },
      'view_distance': 20
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('chart/options_3d/01.js', False, None),

    ('chart/options_3d/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('chart/options_3d/01.js', True, None),

    ('chart/options_3d/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.options.series.data.cartesian import CartesianData


class BoxPlotData(CartesianData):
    """Variant of :class:`CartesianData` which is used for data points in a boxplot."""

    def __init__(self, **kwargs):
        self._box_dash_style = None
        self._high = None
        self._low = None
        self._median = None
        self._median_dash_style = None
        self._q1 = None
        self._q3 = None
        self._stem_dash_style = None
        self._whisker_dash_style = None

        self.box_dash_style = kwargs.get('box_dash_style', None)
        self.high = kwargs.get('high', None)
        self.low = kwargs.get('low', None)
        self.median = kwargs.get('median', None)
        self.median_dash_style = kwargs.get('median_dash_style', None)
        self.q1 = kwargs.get('q1', None)
        self.q3 = kwargs.get('q3', None)
        self.stem_dash_style = kwargs.get('stem_dash_style', None)
        self.whisker_dash_style = kwargs.get('whisker_dash_style', None)

        super().__init__(**kwargs)

    @property
    def box_dash_style(self) -> Optional[str]:
        """The dash style of the box.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._box_dash_style

    @box_dash_style.setter
    def box_dash_style(self, value):
        if not value:
            self._box_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'box_dash_style expects a recognized '
                                                  f'value, but received: {value}')
            self._box_dash_style = value

    @property
    def high(self) -> Optional[int | float | Decimal]:
        """The highest value in the sample set. The top whisker is drawn here. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._high

    @high.setter
    def high(self, value):
        self._high = validators.numeric(value, allow_empty = True)

    @property
    def low(self) -> Optional[int | float | Decimal]:
        """The lowest value in the sample set. The bottom whisker is drawn here. Defaults
        to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._low

    @low.setter
    def low(self, value):
        self._low = validators.numeric(value, allow_empty = True)

    @property
    def median(self) -> Optional[int | float | Decimal]:
        """The median value in the sample set. This is drawn as a line through the middle
        of the box. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._median

    @median.setter
    def median(self, value):
        self._median = validators.numeric(value, allow_empty = True)

    @property
    def median_dash_style(self) -> Optional[str]:
        """The dash style of the median. Defaults to ``'Solid'``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._median_dash_style

    @median_dash_style.setter
    def median_dash_style(self, value):
        if not value:
            self._median_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'median_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._median_dash_style = value

    @property
    def q1(self) -> Optional[int | float | Decimal]:
        """The lower quartile in the sample set. This is the bottom of the box. Defaults
        to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._q1

    @q1.setter
    def q1(self, value):
        self._q1 = validators.numeric(value, allow_empty = True)

    @property
    def q3(self) -> Optional[int | float | Decimal]:
        """The higher quartile in the sample set. This is the top of the box. Defaults
        to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._q3

    @q3.setter
    def q3(self, value):
        self._q3 = validators.numeric(value, allow_empty = True)

    @property
    def stem_dash_style(self) -> Optional[str]:
        """The dash style of the :term:`stem`, the vertical line extending from the box to
        the whiskers. Defaults to ``'Solid'``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stem_dash_style

    @stem_dash_style.setter
    def stem_dash_style(self, value):
        if not value:
            self._stem_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'stem_dash_style expects a recognized'
                                                  f' value, but received: {value}')
            self._stem_dash_style = value

    @property
    def whisker_dash_style(self) -> Optional[str]:
        """The dash style of the whiskers. Defaults to ``'Solid'``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._whisker_dash_style

    @whisker_dash_style.setter
    def whisker_dash_style(self, value):
        if not value:
            self._whisker_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'whisker_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._whisker_dash_style = value

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'BoxPlotData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            elif checkers.is_iterable(item):
                if len(item) == 6:
                    as_dict = {
                        'x': item[0],
                        'low': item[1],
                        'q1': item[2],
                        'median': item[3],
                        'q3': item[4],
                        'high': item[5]
                    }
                elif len(item) == 5:
                    as_dict = {
                        'x': None,
                        'low': item[0],
                        'q1': item[1],
                        'median': item[2],
                        'q3': item[3],
                        'high': item[4]
                    }
                else:
                    raise errors.HighchartsValueError(f'data expects either a 6D or 5D '
                                                      f'collection. Collection received '
                                                      f'had {len(item)} dimensions.')

                as_obj = cls.from_dict(as_dict)
                if checkers.is_string(as_obj.x):
                    as_obj.name = as_obj.x
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a BoxPlot Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank', None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'marker': as_dict.get('marker', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            'box_dash_style': as_dict.get('boxDashStyle', None),
            'high': as_dict.get('high', None),
            'low': as_dict.get('low', None),
            'median': as_dict.get('median', None),
            'median_dash_style': as_dict.get('medianDashStyle', None),
            'q1': as_dict.get('q1', None),
            'q3': as_dict.get('q3', None),
            'stem_dash_style': as_dict.get('stemDashStyle', None),
            'whisker_dash_style': as_dict.get('whiskerDashStyle', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'boxDashStyle': self.box_dash_style,
            'high': self.high,
            'low': self.low,
            'median': self.median,
            'medianDashStyle': self.median_dash_style,
            'q1': self.q1,
            'q3': self.q3,
            'stemDashStyle': self.stem_dash_style,
            'whiskerDashStyle': self.whisker_dash_style,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors, constants
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.utility_functions import mro__to_untrimmed_dict
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.options.sonification.mapping import SonificationMapping
from highcharts_core.options.sonification.grouping import SonificationGrouping


class ActiveWhen(HighchartsMeta):
    """Definition of the condition for when a track should be active or not."""
    
    def __init__(self, **kwargs):
        self._crossing_down = None
        self._crossing_up = None
        self._max = None
        self._min = None
        self._prop = None
        
        self.crossing_down = kwargs.get('crossing_down', None)
        self.crossing_up = kwargs.get('crossing_up', None)
        self.max = kwargs.get('max', None)
        self.min = kwargs.get('min', None)
        self.prop = kwargs.get('prop', None)
        
    @property
    def crossing_down(self) -> Optional[int | float | Decimal]:
        """Track will be active when the property indicated by
        :meth:`.prop <highcharts_core.options.sonification.track_configurations.ActiveWhen.prop>` is
        at or *below* this value. Defaults to :obj:`None <python:None>`.
        
        .. warning::
        
          If both ``.crossing_down`` and 
          :meth:`.crossing_up <highcharts_core.options.sonification.track_configurations.ActiveWhen.crossing_up>` are 
          defined, the track will be active if either condition is met.
          
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._crossing_down
    
    @crossing_down.setter
    def crossing_down(self, value):
        self._crossing_down = validators.numeric(value, allow_empty = True)
        
    @property
    def crossing_up(self) -> Optional[int | float | Decimal]:
        """Track will be active when the property indicated by
        :meth:`.prop <highcharts_core.options.sonification.track_configurations.ActiveWhen.prop>` is
        at or *above* this value. Defaults to :obj:`None <python:None>`.
        
        .. warning::
        
          If both ``.crossing_up`` and 
          :meth:`.crossing_down <highcharts_core.options.sonification.track_configurations.ActiveWhen.crossing_down>` 
          are defined, the track will be active if either condition is met.
          
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._crossing_up
    
    @crossing_up.setter
    def crossing_up(self, value):
        self._crossing_up = validators.numeric(value, allow_empty = True)
    
    @property
    def max(self) -> Optional[int | float | Decimal]:
        """Track will be active when the property indicated by
        :meth:`.prop <highcharts_core.options.sonification.track_configurations.ActiveWhen.prop>` is
        at or *below* this value. Defaults to :obj:`None <python:None>`.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max
    
    @max.setter
    def max(self, value):
        self._max = validators.numeric(value, allow_empty = True)
        
    @property
    def min(self) -> Optional[int | float | Decimal]:
        """Track will be active when the property indicated by
        :meth:`.prop <highcharts_core.options.sonification.track_configurations.ActiveWhen.prop>` is
        at or *above* this value. Defaults to :obj:`None <python:None>`.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min
    
    @min.setter
    def min(self, value):
        self._min = validators.numeric(value, allow_empty = True)
        
    @property
    def prop(self) -> Optional[str]:
        """The data point property to use when evaluating the condition, for example ``'y'`` or ``'x'``.
        Defaults to :obj:`None <python:None>`.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._prop
    
    @prop.setter
    def prop(self, value):
        self._prop = validators.string(value, allow_empty = True)
        
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'crossing_down': as_dict.get('crossingDown', None),
            'crossing_up': as_dict.get('crossingUp', None),
            'max': as_dict.get('max', None),
            'min': as_dict.get('min', None),
            'prop': as_dict.get('prop', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'crossingDown': self.crossing_down,
            'crossingUp': self.crossing_up,
            'max': self.max,
            'min': self.min,
            'prop': self.prop,
        }

        return untrimmed


class TrackConfigurationBase(HighchartsMeta):
    """Base class for use in configuring Sonification tracks."""
    
    def __init__(self, **kwargs):
        self._active_when = None
        self._mapping = None
        self._midi_name = None
        self._point_grouping = None
        self._show_play_marker = None
        self._type = None
        
        self.active_when = kwargs.get('active_when', None)
        self.mapping = kwargs.get('mapping', None)
        self.midi_name = kwargs.get('midi_name', None)
        self.point_grouping = kwargs.get('point_grouping', None)
        self.show_play_marker = kwargs.get('show_play_marker', None)
        self.type = kwargs.get('type', None)
        
    @property
    def active_when(self) -> Optional[ActiveWhen | CallbackFunction]:
        """The condition for when a track should be active or not.
        
        Accepts either a (Javascript) 
        :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or an
        :class:`ActiveWhen <highcharts_core.options.sonification.track_configurations.ActiveWhen>` configuration object.
        
        .. note::

          If a callback function is used, it should return a boolean for whether or not the track should be active. 
          
          The function is called for each audio event, and receives a parameter object with ``time``, and potentially 
          ``point`` and ``value`` properties depending on the track. 
          
          ``point`` is available if the audio event is related to a data point. 
          
          ``value`` is available if the track is used as a context track, and 
          :meth:`.value_interval <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration.value_interval>` 
          is used.
          
        :rtype: :class:`ActiveWhen <highcharts_core.options.sonification.track_configurations.ActiveWhen>` or
          :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`.
        """
        return self._active_when
    
    @active_when.setter
    def active_when(self, value):
        if not value:
            self._active_when = None
        else:
            try:
                value = validate_types(value, types = (ActiveWhen))
            except (ValueError, TypeError):
                value = validate_types(value, types = (CallbackFunction))
            
            self._active_when = value

    @property
    def mapping(self) -> Optional[SonificationMapping]:
        """Mapping options for the audio parameter.
        
        :rtype: :class:`SonificationMapping <highcharts_core.options.sonification.mapping.SonificationMapping>` or
          :obj:`None <python:None>`
        """
        return self._mapping
    
    @mapping.setter
    @class_sensitive(SonificationMapping)
    def mapping(self, value):
        self._mapping = value
        
    @property
    def point_grouping(self) -> Optional[SonificationGrouping]:
        """Options for configurign the grouping of points.
        
        :rtype: :class:`SonificationGrouping <highcharts_core.options.sonification.grouping.SonificationGrouping>` or
          :obj:`None <python:None>`
        """
        return self._point_grouping
    
    @point_grouping.setter
    @class_sensitive(SonificationGrouping)
    def point_grouping(self, value):
        self._point_grouping = value
        
    @property
    def show_play_marker(self) -> Optional[bool]:
        """If ``True``, displays the play marker (tooltip and/or crosshair) for a track. Defaults to ``True``.
        
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_play_marker
    
    @show_play_marker.setter
    def show_play_marker(self, value):
        if value is None:
            self._show_play_marker = None
        else:
            self._show_play_marker = bool(value)
            
    @property
    def type(self) -> Optional[str]:
        """The type of track. Accepts either ``'instrument'`` or ``'speech'``. Defaults to ``'instrument'``.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type
    
    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['instrument', 'speech']:
                raise errors.HighchartsValueError(f'type expects either "instrument" or "speech". '
                                                  f'Received: "{value}"')
            self._type = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'active_when': as_dict.get('activeWhen', None),
            'mapping': as_dict.get('mapping', None),
            'midi_name': as_dict.get('midiName', None) or as_dict.get('MIDIName', None),
            'point_grouping': as_dict.get('pointGrouping', None),
            'show_play_marker': as_dict.get('showPlayMarker', None),
            'type': as_dict.get('type', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'activeWhen': self.active_when,
            'mapping': self.mapping,
            'midiName': self.midi_name,
            'pointGrouping': self.point_grouping,
            'showPlayMarker': self.show_play_marker,
            'type': self.type,
        }

        return untrimmed


class InstrumentTrackConfiguration(TrackConfigurationBase):
    """Configuration of an Instrument Track for use in sonification."""
    
    def __init__(self, **kwargs):
        self._instrument = None
        self._round_to_musical_notes = None
        
        self.instrument = kwargs.get('instrument', None)
        self.round_to_musical_notes = kwargs.get('round_to_musical_notes', None)
        
        super().__init__(**kwargs)
        
    @property
    def instrument(self) -> Optional[str]:
        """The instrument to use for playing. Defaults to ``'piano'``.
        
        Accepts:

          * ``'flute'``
          * ``'saxophone'``
          * ``'trumpet'``
          * ``'sawsynth'``
          * ``'wobble'``
          * ``'basic1'``
          * ``'basic2'``
          * ``'sine'``
          * ``'sineGlide'``
          * ``'triangle'``
          * ``'square'``
          * ``'sawtooth'``
          * ``'noise'``
          * ``'filteredNoise'``
          * ``'wind'``
          
        :rtype: :class:`str <python:str>`
        """
        return self._instrument
    
    @instrument.setter
    def instrument(self, value):
        if not value:
            self._instrument = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in constants.INSTRUMENT_PRESETS:
                raise errors.HighchartsValueError(f'.instrument expects a predefined instrument name. Did not '
                                                  f'recognize: "{value}".')
            self._instrument = value

    @property
    def midi_name(self) -> Optional[str]:
        """The name to use for a track when exporting it to MIDI. If :obj:`None <python:None>`, will
        use the series name if the track is related to a series. Defaults to :obj:`None <python:None>`.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._midi_name
    
    @midi_name.setter
    def midi_name(self, value):
        self._midi_name = validators.string(value, allow_empty = True)
        
    @property
    def round_to_musical_notes(self) -> Optional[bool]:
        """If ``True``, will round pitch matching to musical notes in 440Hz standard tuning. If ``False``,
        will play the exact mapped/configured note even if it is out of tune as per standard tuning. Defaults to ``True``.
        
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._round_to_musical_notes
    
    @round_to_musical_notes.setter
    def round_to_musical_notes(self, value):
        if value is None:
            self._round_to_musical_notes = None
        else:
            self._round_to_musical_notes = bool(value)
            
    @property
    def type(self) -> Optional[str]:
        """The type of track.
        
        .. note::
        
          In the context of an :class:`InstrumentTrackConfiguration`, this will *always* return ``'instrument'`` if
          not :obj:`None <python:None>`.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        if self._instrument:
            return 'instrument'
        
        return None
    
    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['instrument', 'speech']:
                raise errors.HighchartsValueError(f'type expects either "instrument" or "speech". '
                                                  f'Received: "{value}"')
            self._type = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'active_when': as_dict.get('activeWhen', None),
            'mapping': as_dict.get('mapping', None),
            'point_grouping': as_dict.get('pointGrouping', None),
            'show_play_marker': as_dict.get('showPlayMarker', None),
            'type': as_dict.get('type', None),
            
            'instrument': as_dict.get('instrument', None),
            'midi_name': as_dict.get('midiName', None) or as_dict.get('MIDIName', None),
            'round_to_musical_notes': as_dict.get('roundToMusicalNotes', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'instrument': self.instrument,
            'midiName': self.midi_name,
            'roundToMusicalNotes': self.round_to_musical_notes,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class SpeechTrackConfiguration(TrackConfigurationBase):
    """Configuration of a Speech Track for use in sonification."""
    
    def __init__(self, **kwargs):
        self._language = None
        self._preferred_voice = None
        
        self.language = kwargs.get('language', None)
        self.preferred_voice = kwargs.get('preferred_voice', None)
        
        super().__init__(**kwargs)
        
    @property
    def language(self) -> Optional[str]:
        """The language to speak in for speech tracks, as an `IETF BCP 47 <https://www.rfc-editor.org/info/bcp47>`__ 
        language tag. Defaults to ``'en-US'``.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._language
    
    @language.setter
    def language(self, value):
        self._language = validators.string(value, allow_empty = True)
        
    @property
    def preferred_voice(self) -> Optional[str]:
        """The name of the voice synthesis to prefer for speech tracks. If :obj:`None <python:None>` or
        unavabilable, will fall back to the default voice for the selected language. Defaults to 
        :obj:`None <python:None>`.
          
        .. warning::
        
          Different platforms (operating systems in which your users will view your visualizations)
          provide different voices for web speech synthesis.
          
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._preferred_voice
    
    @preferred_voice.setter
    def preferred_voice(self, value):
        self._preferred_voice = validators.string(value, allow_empty = True)
        
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'active_when': as_dict.get('activeWhen', None),
            'mapping': as_dict.get('mapping', None),
            'point_grouping': as_dict.get('pointGrouping', None),
            'show_play_marker': as_dict.get('showPlayMarker', None),
            'type': as_dict.get('type', None),
            
            'language': as_dict.get('language', None),
            'preferred_voice': as_dict.get('preferredVoice', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'language': self.language,
            'preferredVoice': self.preferred_voice,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class ContextTrackConfiguration(InstrumentTrackConfiguration, SpeechTrackConfiguration):
    """Configuration of a Context Track for use in sonification."""
    
    def __init__(self, **kwargs):
        self._time_interval = None
        self._value_interval = None
        self._value_map_function = None
        self._value_prop = None
        
        self.time_interval = kwargs.get('time_interval', None)
        self.value_interval = kwargs.get('value_interval', None)
        self.value_map_function = kwargs.get('value_map_function', None)
        self.value_prop = kwargs.get('value_prop', None)
        
        super().__init__(**kwargs)
        
    @property
    def time_interval(self) -> Optional[int | float | Decimal]:
        """Determines the number of milliseconds between playback of a context track. Defaults to 
        :obj:`None <python:None>`.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._time_interval
    
    @time_interval.setter
    def time_interval(self, value):
        self._time_interval = validators.numeric(value, allow_empty = True)

    @property
    def value_interval(self) -> Optional[int | float | Decimal]:
        """Determines the number of units between playback of a context track, where
        units are determined by :meth:`.value_prop <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration.value_prop>`.
        
        For example, setting 
        :meth:`.value_prop <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration.value_prop>` 
        to ``'x'`` and ``.value_interval`` to ``5`` means the context track
        should be played for every 5th value of ``'x'``.
        
        .. note::
        
          The context audio events will be mapped to time according to the prop value relative to the min/max values 
          for that prop.
          
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value_interval
    
    @value_interval.setter
    def value_interval(self, value):
        self._value_interval = validators.numeric(value, allow_empty = True)
        
    @property
    def value_map_function(self) -> Optional[str]:
        """Determines how to map context events to time when using the 
        :meth:`.value_interval <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration.value_interval>`
        property. Accepts either ``'linear'`` or ``'logarithmic'``. Defaults to ``'linear'``.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_map_function
    
    @value_map_function.setter
    def value_map_function(self, value):
        if not value:
            self._value_map_function = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['linear', 'logarithmic']:
                raise errors.HighchartsValueError(f'value_map_function expects either "linear" or '
                                                  f'"logarithmic. Received: "{value}"')
            
            self._value_map_function = value
            
    @property
    def value_prop(self) -> Optional[str]:
        """The data point property to use when evaluating whether to play the context track in conjunction with 
        :meth:`.value_interval <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration.value_interval>`
        Defaults to :obj:`None <python:None>`.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_prop
    
    @value_prop.setter
    def value_prop(self, value):
        self._value_prop = validators.string(value, allow_empty = True)

    @property
    def type(self) -> Optional[str]:
        """The type of track. Accepts either ``'instrument'`` or ``'speech'``. Defaults to ``'instrument'``.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['instrument', 'speech']:
                raise errors.HighchartsValueError(f'type expects either "instrument" or "speech". '
                                                  f'Received: "{value}"')
            self._type = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'active_when': as_dict.get('activeWhen', None),
            'mapping': as_dict.get('mapping', None),
            'point_grouping': as_dict.get('pointGrouping', None),
            'show_play_marker': as_dict.get('showPlayMarker', None),
            'type': as_dict.get('type', None),
            
            'instrument': as_dict.get('instrument', None),
            'midi_name': as_dict.get('midiName', None) or as_dict.get('MIDIName', None),
            'round_to_musical_notes': as_dict.get('roundToMusicalNotes', None),

            'language': as_dict.get('language', None),
            'preferred_voice': as_dict.get('preferredVoice', None),
            
            'time_interval': as_dict.get('timeInterval', None),
            'value_interval': as_dict.get('valueInterval', None),
            'value_map_function': as_dict.get('valueMapFunction', None),
            'value_prop': as_dict.get('valueProp', None),
            
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'timeInterval': self.time_interval,
            'valueInterval': self.value_interval,
            'valueMapFunction': self.value_map_function,
            'valueProp': self.value_prop,
        }

        parent_as_dict = mro__to_untrimmed_dict(self, in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


from typing import Optional, List

from validator_collection import validators

from highcharts_core.decorators import validate_types, class_sensitive
from highcharts_core.utility_classes.ast import AttributeObject, ASTNode, ASTMap


class MarkerAttributeObject(AttributeObject):
    """Special subclass of the :class:`SVGAttributeMap` which provides properties for
    annotation markers.

    While this class also is a special subclass of :class:`dict <python:dict>`, it also
    propagates values from certain named properties to corresponding key entries in the
    underling :class:`dict <python:dict>`'s data map.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self['id'] = None
        self['refX'] = None
        self['refY'] = None
        self['markerWidth'] = None
        self['markerHeight'] = None

        self.id = kwargs.get('id', None)
        self.ref_x = kwargs.get('ref_x', None)
        self.ref_y = kwargs.get('ref_y', None)
        self.marker_width = kwargs.get('marker_width', None)
        self.marker_height = kwargs.get('marker_height', None)

    @property
    def id(self) -> Optional[str]:
        """The identifier applied to the marker.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self['id']

    @id.setter
    def id(self, value):
        value = validators.string(value, allow_empty = True)
        self['id'] = value

    @property
    def ref_x(self) -> Optional[int]:
        """TBD

        .. todo::

          Determine what purpose the refX and refY properties serve.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self.get('refX')

    @ref_x.setter
    def ref_x(self, value):
        value = validators.integer(value, allow_empty = True, coerce_value = True)
        self['refX'] = value

    @property
    def ref_y(self) -> Optional[int]:
        """TBD

        .. todo::

          Determine what purpose the refX and refY properties serve.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self['refY']

    @ref_y.setter
    def ref_y(self, value):
        value = validators.integer(value, allow_empty = True, coerce_value = True)
        self['refY'] = value

    @property
    def marker_width(self) -> Optional[int]:
        """The width of the marker, expressed in pixels.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self['markerWidth']

    @marker_width.setter
    def marker_width(self, value):
        value = validators.integer(value, allow_empty = True, coerce_value = True)
        self['markerWidth'] = value

    @property
    def marker_height(self) -> Optional[int]:
        """The height of the marker, expressed in pixels.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self['markerHeight']

    @marker_height.setter
    def marker_height(self, value):
        value = validators.integer(value, allow_empty = True, coerce_value = True)
        self['markerHeight'] = value


class MarkerASTNode(ASTNode):
    """:class:`ASTNode` that enforces a Marker Attribute set.."""

    @property
    def attributes(self) -> Optional[MarkerAttributeObject]:
        """Map of attributes that are used to describe the node.

        Basically corresponds to a :class:`dict <python:dict>`, where keys have to
        generally conform to standard variable-naming conventions and values are basic
        literals (string, array, numer, boolean, :class:`dict <python:dict>`, etc.)

        :rtype: :class:`MarkerAttributeObject` or :obj:`None <python:None>`
        """
        return self._attributes

    @attributes.setter
    @class_sensitive(MarkerAttributeObject)
    def attributes(self, value):
        self._attributes = value

    @property
    def children(self) -> Optional[List[ASTNode]]:
        """Collection of :class:`ASTNode` instances that should be considered children
        of this node.

        :rtype: :class:`list <python:list>` of :class:`ASTNode` objects
        """
        return self._children

    @children.setter
    def children(self, value):
        if not value:
            self._children = []
        else:
            value = validate_types(value,
                                   types = ASTNode,
                                   allow_none = False,
                                   force_iterable = True)
            self._children = value


class MarkerDefinition(ASTMap):
    """Special :class:`dict <python:dict>` class used to construct a JavaScript object
    literal map. Each key represents a key or node within that map, while
    the value for that key is an :class:`ASTNode`.

    Keys are validated to be valid variable names, while each value is validated to be an
    :class:`ASTNode`.

    """

    def __setitem__(self, key, item):
        if key not in self:
            key = validators.variable_name(key, allow_empty = False)

        item = validate_types(item,
                              types = MarkerASTNode,
                              allow_none = False)

        super().__setitem__(key, item)



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class BubbleLegendLabelOptions(HighchartsMeta):
    """Options to configure the bubble legend's labels."""

    def __init__(self, **kwargs):
        self._align = None
        self._allow_overlap = None
        self._class_name = None
        self._format = None
        self._formatter = None
        self._style = None
        self._x = None
        self._y = None

        self.align = kwargs.get('align', None)
        self.allow_overlap = kwargs.get('allow_overlap', None)
        self.class_name = kwargs.get('class_name', None)
        self.format = kwargs.get('format', None)
        self.formatter = kwargs.get('formatter', None)
        self.style = kwargs.get('style', None)
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'legend.bubbleLegend'

    @property
    def align(self) -> Optional[str]:
        """The alignment of the labels relative to the bubble legend. Defaults to
        ``'right'``.

        Accepts:

          * ``'left'``
          * ``'center'``
          * ``'right'``

        :returns: The alignment of the annotation's label.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._align

    @align.setter
    def align(self, value):
        if not value:
            self._align = None
        else:
            value = validators.string(value, allow_empty = True)
            value = value.lower()
            if value not in ['left', 'center', 'right']:
                raise errors.HighchartsValueError(f'align must be either "left", "center"'
                                                  f', or "right". Was: {value}')

        self._align = value

    @property
    def allow_overlap(self) -> Optional[bool]:
        """If ``True``, data labels are allowed to overlap each other.

        Defaults to ``False``.

        :returns: Flag indicating whether to allow data labels to overlap.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_overlap

    @allow_overlap.setter
    def allow_overlap(self, value):
        if value is None:
            self._allow_overlap = None
        else:
            self._allow_overlap = bool(value)

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS. Defaults to
        ``'None'``.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def format(self) -> Optional[str]:
        """A format string to apply to the label.

        :returns: The format string to apply to the labels.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the bubble legend's data labels.

        .. hint::

          In the JavaScript callback function, the ``this`` properties available are:

            * ``this.value`` - the bubble value
            * ``this.radius`` - the bubble radius
            * ``this.center`` - the y position of the bubble's center

        :returns: A JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._formatter

    @formatter.setter
    @class_sensitive(CallbackFunction)
    def formatter(self, value):
        self._formatter = value

    @property
    def style(self) -> Optional[str]:
        """CSS styling to apply to the data labels.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def x(self) -> Optional[int]:
        """The x position offset of the label relative to the connector. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def y(self) -> Optional[int]:
        """The y position offset of the label relative to the connector. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'class_name': as_dict.get('className', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'style': as_dict.get('style', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'align': self.align,
            'allowOverlap': self.allow_overlap,
            'className': self.class_name,
            'format': self.format,
            'formatter': self.formatter,
            'style': self.style,
            'x': self.x,
            'y': self.y
        }

        return untrimmed


class BubbleLegendRange(HighchartsMeta):
    """Options for specific range. One range consists of bubble, label and connector."""

    def __init__(self, **kwargs):
        self._border_color = None
        self._color = None
        self._connector_color = None
        self._value = None

        self.border_color = kwargs.get('border_color', None)
        self.color = kwargs.get('color', None)
        self.connector_color = kwargs.get('connector_color', None)
        self.value = kwargs.get('value', None)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the range's border. Defaults to :obj:`None <python:None>`.

        :returns: The color of the range borders.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The main color of the bubble for the range. Defaults to
        :obj:`None <python:None>`.

        :returns: The main color of the bubble legend.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def connector_color(self) -> Optional[str | Gradient | Pattern]:
        """The color applied to the connector for the range. Defaults to
        :obj:`None <python:None>`.

        :returns: The color applied to the connector.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._connector_color

    @connector_color.setter
    def connector_color(self, value):
        from highcharts_core import utility_functions
        self._connector_color = utility_functions.validate_color(value)

    @property
    def value(self) -> Optional[int | float | Decimal]:
        """The range size value, similar to the bubble's Z-value. Defaults to
        :obj:`None <python:None>`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value_):
        self._value = validators.numeric(value_, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'color': as_dict.get('color', None),
            'connector_color': as_dict.get('connectorColor', None),
            'value': as_dict.get('value', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderColor': self.border_color,
            'color': self.color,
            'connectorColor': self.connector_color,
            'value': self.value
        }

        return untrimmed


class BubbleLegend(HighchartsMeta):
    """The bubble legend is an additional element in legend which presents the scale
    of the bubble series.

    Individual bubble ranges can be defined by user or calculated from series. In the
    case of automatically calculated ranges, a 1px margin of error is permitted.

    """

    def __init__(self, **kwargs):
        self._border_color = None
        self._border_width = None
        self._class_name = None
        self._color = None
        self._connector_class_name = None
        self._connector_color = None
        self._connector_distance = None
        self._connector_width = None
        self._enabled = None
        self._labels = None
        self._legend_index = None
        self._max_size = None
        self._min_size = None
        self._ranges = None
        self._size_by = None
        self._size_by_absolute_value = None
        self._z_index = None
        self._z_threshold = None

        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.connector_class_name = kwargs.get('connector_class_name', None)
        self.connector_color = kwargs.get('connector_color', None)
        self.connector_distance = kwargs.get('connector_distance', None)
        self.connector_width = kwargs.get('connector_width', None)
        self.enabled = kwargs.get('enabled', None)
        self.labels = kwargs.get('labels', None)
        self.legend_index = kwargs.get('legend_index', None)
        self.max_size = kwargs.get('max_size', None)
        self.min_size = kwargs.get('min_size', None)
        self.ranges = kwargs.get('ranges', None)
        self.size_by = kwargs.get('size_by', None)
        self.size_by_absolute_value = kwargs.get('size_by_absolute_value', None)
        self.z_index = kwargs.get('z_index', None)
        self.z_threshold = kwargs.get('z_threshold', None)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the ranges border. Can also be defined for an individual range.
        Defaults to :obj:`None <python:None>`.

        :returns: The color of the range borders.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the range borders. Can also be defined
        for an individual range. Defaults to :obj:`None <python:None>`

        :returns: The border width to apply to the range borders.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The main color of the bubble legend. Applies to ranges, if individual ranges
        are not given a color. Defaults to :obj:`None <python:None>`.

        :returns: The main color of the bubble legend.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def connector_class_name(self) -> Optional[str]:
        """An additional class name to apply to the bubble legend's connector graphical
        elements. This option does not replace default class names of the graphical
        element. Defaults to :obj:`None <python:None>`.

        :returns: The class name applied to the bubble legend's connector graphical
          elements.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._connector_class_name

    @connector_class_name.setter
    def connector_class_name(self, value):
        self._connector_class_name = validators.string(value, allow_empty = True)

    @property
    def connector_color(self) -> Optional[str | Gradient | Pattern]:
        """The color applied to the connector. Can also be defined for individual ranges.
        Defaults to :obj:`None <python:None>`.

        :returns: The color applied to the connector.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._connector_color

    @connector_color.setter
    def connector_color(self, value):
        from highcharts_core import utility_functions
        self._connector_color = utility_functions.validate_color(value)

    @property
    def connector_distance(self) -> Optional[int | float | Decimal]:
        """The length of the connector in pixels. Defaults to
        ``60``.

        .. note::

          If labels are centered, the distance is automatically reduced to ``0``.

        :returns: The border width to apply to the range borders.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._connector_distance

    @connector_distance.setter
    def connector_distance(self, value):
        self._connector_distance = validators.numeric(value, allow_empty = True)

    @property
    def connector_width(self) -> Optional[int | float | Decimal]:
        """The width of the connector in pixels. Defaults to
        ``1``.

        :returns: The border width to apply to the range borders.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._connector_width

    @connector_width.setter
    def connector_width(self, value):
        self._connector_width = validators.numeric(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, displays the bubble legend. If ``False``, hides the legend.
        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def labels(self) -> Optional[BubbleLegendLabelOptions]:
        """Options to configure the bubble legend's labels. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`BubbleLegendLabelOptions` or :obj:`None <python:None>`
        """
        return self._labels

    @labels.setter
    @class_sensitive(BubbleLegendLabelOptions)
    def labels(self, value):
        self._labels = value

    @property
    def legend_index(self) -> Optional[int]:
        """The position of the bubble legend within the legend. Defaults to
        ``0``.

        :returns: The position of the bubble legend within the legend.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._legend_index

    @legend_index.setter
    def legend_index(self, value):
        self._legend_index = validators.integer(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def max_size(self) -> Optional[int | float | Decimal]:
        """The maximum bubble legend range size, in pixels. Defaults to
        ``60``.

        .. note::

          If not specified, the maximum size is determined automatically.

        :returns: The maximum bubble legend range size.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_size

    @max_size.setter
    def max_size(self, value):
        self._max_size = validators.numeric(value, allow_empty = True)

    @property
    def min_size(self) -> Optional[int | float | Decimal]:
        """The minimum bubble legend range size, in pixels. Defaults to
        ``10``.

        .. note::

          If not specified, the minimum size is determined automatically.

        :returns: The minimum bubble legend range size.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_size

    @min_size.setter
    def min_size(self, value):
        self._min_size = validators.numeric(value, allow_empty = True)

    @property
    def ranges(self) -> Optional[List[BubbleLegendRange]]:
        """Options for specific range. One range consists of bubble, label and connector.

        Defaults to :obj:`None <python:None>`

        :rtype: :class:`list <python:list>` of :class:`BubbleLegendRange`, or
          :obj:`None <python:None>`
        """
        return self._ranges

    @ranges.setter
    @class_sensitive(BubbleLegendRange, force_iterable = True)
    def ranges(self, value):
        self._ranges = value

    @property
    def size_by(self) -> Optional[str]:
        """Indicates whether the bubble legend range should be represented by the area
        or the width of the bubble. The default
        (``'area'``) corresponds best to the
        human perception of the size of each bubble.

        Accepts one of two possible values:

          * ``'area'``
          * ``'width'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._size_by

    @size_by.setter
    def size_by(self, value):
        if not value:
            self._size_by = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['area', 'width']:
                raise errors.HighchartsValueError(f'size_by expects either "area" or '
                                                  f'"width". Received: {value}')
            self._size_by = value

    @property
    def size_by_absolute_value(self) -> Optional[bool]:
        """If ``True``, the absolute value of z determines the size of the bubble. This
        means that with the default :meth:`BubbleLegend.z_threshold` of ``0``, a bubble of
        value ``-1`` will have the same size as a bubble of value ``1``, while a bubble of
        value 0 will have a smaller size according to :meth:`BubbleLegend.min_size`.

        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._size_by_absolute_value

    @size_by_absolute_value.setter
    def size_by_absolute_value(self, value):
        if value is None:
            self._size_by_absolute_value = None
        else:
            self._size_by_absolute_value = bool(value)

    @property
    def z_index(self) -> Optional[int]:
        """The visual z index of the bubble legend. Defaults to
        ``1``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.integer(value, allow_empty = True)

    @property
    def z_threshold(self) -> Optional[int | float | Decimal]:
        """Ranges with a lower z-value are skipped in the legend. Defaults to
        ``0``.

        :rtype: numeric
        """
        return self._z_threshold

    @z_threshold.setter
    def z_threshold(self, value):
        self._z_threshold = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'connector_class_name': as_dict.get('connectorClassName', None),
            'connector_color': as_dict.get('connectorColor', None),
            'connector_distance': as_dict.get('connectorDistance', None),
            'connector_width': as_dict.get('connectorWidth', None),
            'enabled': as_dict.get('enabled', None),
            'labels': as_dict.get('labels', None),
            'legend_index': as_dict.get('legendIndex', None),
            'max_size': as_dict.get('maxSize', None),
            'min_size': as_dict.get('minSize', None),
            'ranges': as_dict.get('ranges', None),
            'size_by': as_dict.get('sizeBy', None),
            'size_by_absolute_value': as_dict.get('sizeByAbsoluteValue', None),
            'z_index': as_dict.get('zIndex', None),
            'z_threshold': as_dict.get('zThreshold', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'className': self.class_name,
            'color': self.color,
            'connectorClassName': self.connector_class_name,
            'connectorColor': self.connector_color,
            'connectorDistance': self.connector_distance,
            'connectorWidth': self.connector_width,
            'enabled': self.enabled,
            'labels': self.labels,
            'legendIndex': self.legend_index,
            'maxSize': self.max_size,
            'minSize': self.min_size,
            'ranges': self.ranges,
            'sizeBy': self.size_by,
            'sizeByAbsoluteValue': self.size_by_absolute_value,
            'zIndex': self.z_index,
            'zThreshold': self.z_threshold
        }

        return untrimmed
